; loader.S
;
#include "lilo.h"

	segment	.text

	global	First
	global	Second
	global	Third
	global	Bitmap
	global	Mbr
	global	Chain
	global	Os2_d
	global	Mbr2

First		dd	FirstEnd-$-4
#ifdef LCF_BUILTIN
incbin "first.b"
#endif
FirstEnd	equ	$

Second		dd	SecondEnd-$-4
#ifdef LCF_BUILTIN
incbin "second.b"
#endif
SecondEnd	equ	$

Third		dd	ThirdEnd-$-4
#ifdef LCF_BUILTIN
incbin	"third.b"
#endif
ThirdEnd	equ	$

Bitmap		dd	BitmapEnd-$-4
#ifdef LCF_BUILTIN
incbin	"bitmap.b"
#endif
BitmapEnd	equ	$

Mbr		dd	MbrEnd-$-4
#ifdef LCF_BUILTIN
incbin	"mbr.b"
#endif
MbrEnd		equ	$

Mbr2		dd	Mbr2End-$-4
#ifdef LCF_BUILTIN
incbin	"mbr2.b"
#endif
Mbr2End		equ	$

Chain		dd	ChainEnd-$-4
#ifdef LCF_BUILTIN
incbin	"chain.b"
#endif
ChainEnd	equ	$

#ifndef LCF_SOLO_CHAIN
Os2_d		dd	Os2_dEnd-$-4
#ifdef LCF_BUILTIN
incbin	"os2_d.b"
#endif
#else
Os2_d		equ	Chain
#endif
Os2_dEnd	equ	$
		dd	$-$$

